;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

#|
(enved (list 0 1 10 .25  25 .1 100 0))
|#

(in-package :stella)

(defun draw-line (view xy1 xy2 &key thickness erase?)
  (ccl:with-pen-saved 
    (ccl:set-pen-size view (xy thickness thickness))
    (when erase? (ccl:set-pen-pattern view ccl:*white-pattern*))
    (#_MoveTo :long xy1)
    (#_LineTo :long xy2)))

(defmacro draw-line* (view x1 y1 x2 y2 &rest keys)
  `(draw-line ,view (xy (floor ,x1) (floor ,y1))
              (xy (floor ,x2) (floor ,y2)) ,@ keys))

(defun draw-circle* (view x y &key (radius 1) filled erase?)
  (setf x (floor x) y (floor y))
  (let* ((diam (round radius))
         (rad (/ diam 2))
         (r- (floor rad))
         (r+ (ceiling rad))
         (left (- x r-))       
         (top (- y r+))
         (right (+ x r+))
         (bottom (+ y r-)))
    (if filled
      (ccl:fill-oval view (if erase? ccl:*white-pattern* 
                              ccl:*black-pattern*)
                     left top right bottom)
      (ccl:frame-oval view left top right bottom))))

(defun draw-string* (view string x y &key (align-x ':left) 
                          (align-y ':bottom) transformation)
  ;; align-x = :left :center :right
  ;; align-y = :top :center :bottom
  (declare (ignore transformation))
  (let ((se (string-extent string (ccl:view-font view))))
    (setf x (floor x) y (floor y))
    (ecase align-x
      (:left )
      (:center (setf x (floor (- x (/ (xy-x se) 2)))))
      (:right (setf x (- x (xy-x se)))))
    (ecase align-y
      (:top (setf y (+ y (xy-y se))))
      (:center (setf y (floor (+ y (/ (xy-y se) 2)))))
      (:bottom ))
    (ccl:with-font-focused-view view
      (ccl:with-pstrs ((s string))
        (#_MoveTo x y)
        (#_DrawString s)))))

;;;
;;;
;;;

(defparameter *tick-space* 15)
(defparameter *small-tick* (floor *tick-space*  4))
(defparameter *large-tick* (floor *tick-space* 2))

(defstruct (plot (:type list))
  (name "Untitled")
  coords
  x-axis
  y-axis
  position
  size
  (pen 1)
  (radius 3)
  external)

(defstruct (axis (:type list))
  name
  minimum
  maximum
  labels
  (divisions 1)
  (ticks-per-label 1)
  (decimals 2))

(defmacro coord (x y) `(cons ,x ,y))
(defmacro coord-x (c) `(car ,c))
(defmacro coord-y (c) `(cdr ,c))

(defun parse-coords (list)
  ;; assume external coord representation is a flat list, (x y ...)
  (loop for (x y) on list by #'cddr 
        collect (coord x y)))

(defun unparse-coords (coords)
  ;; return external form
  (loop for c in coords
        collect (coord-x c)
        collect (coord-y c)))

(defun compute-axis-labels (axis)
  (let ((minimum (axis-minimum axis))
        (maximum (axis-maximum axis))
        (divisions (axis-divisions axis))
        (format (format nil "~~,~AF" (axis-decimals axis))))
    (loop with increment = (/ (- maximum minimum) divisions)
          repeat (1+ divisions)
          collect (format nil format minimum)
          do (incf minimum increment))))

(defun draw-x-axis (view x y length labels &key (ticks-per-label 1)
                         (thickness 1) )
  (let* ((divisions (* (1- (length labels)) ticks-per-label))
         (increment (/ length divisions)))
    ;; draw axis.
    (draw-line* view x y (+ x length) y :thickness thickness )
    ;; draw axis ticks and labels.
    (loop for i from 0 to divisions
          if (> (mod i ticks-per-label) 0)
          do
          ;; draw small tick.
          (draw-line* view x y x (+ y *small-tick*) :thickness thickness)
          else do 
          ;; draw large tick and label.
          (draw-line* view x y x (+ y *large-tick*) :thickness thickness)
          (draw-string* view (pop labels) x (+ y *tick-space*)
                        :align-x ':center :align-y ':top)
          do (incf x increment))))

(defun draw-y-axis (view x y height labels &key (ticks-per-label 1)          
                         (thickness 1) )
  (let* ((divisions (* (1- (length labels)) ticks-per-label))
         (increment (- (/ height divisions))))
    ;; draw axis.
    (draw-line* view x y x (- y height) :thickness thickness)
    ;; draw axis ticks and labels.
    (loop for i from 0 to divisions
          if (> (mod i ticks-per-label) 0)    
          do       
          ;; draw small tick.
          (draw-line* view (- x *small-tick*) y x y :thickness thickness)
          else do
          ;; draw large tick and label.
          (draw-line* view (- x *large-tick*) y x y :thickness thickness)
          (let ((label (pop labels)))
            (when label
              (draw-string* view label (- x *tick-space*) y  
                            :align-x ':right :align-y ':center
                            )))
          do (incf y increment))))

(defun draw-coords (view x y length height 
                         x-minimum x-maximum y-minimum y-maximum
                         coords &key (thickness 1) (radius 1)
                         erase?)
  ;; bottom of plot = height.
  (loop with x-range = (- x-maximum x-minimum)
        and  y-range = (- y-maximum y-minimum)
        for coord in coords
        for last-x = nil then this-x
        for this-x = (+ x (* (/ (- (coord-x coord) x-minimum) x-range)
                             length))
        for last-y = nil then this-y
        for this-y = (- y (* (/ (- (coord-y coord) y-minimum) y-range)
                             height))
        do
        (draw-circle* view this-x this-y :radius radius :filled t
                      :erase? erase?)
        (when last-x 
          (draw-line* view last-x last-y this-x this-y 
                      :thickness thickness :erase? erase?))))

(defun draw-plot (view plot &optional (x-axis? t) (y-axis? t) 
                       (coords? t))
  (let ((x-axis (plot-x-axis plot))
        (y-axis (plot-y-axis plot))
        (thickness (plot-pen plot))
        (radius (plot-radius plot))
        (plot-position (plot-position plot))
        (plot-size (plot-size plot))
        plot-x plot-y plot-w plot-h)
    (setf plot-x (xy-x plot-position))
    (setf plot-y (xy-y plot-position))
    (setf plot-w (xy-x plot-size))
    (setf plot-h (xy-y plot-size))
    (when x-axis?
      (draw-x-axis view plot-x plot-y plot-w (axis-labels x-axis)
                   :thickness thickness
                   :ticks-per-label (axis-ticks-per-label x-axis)))  
    (when y-axis?
      (draw-y-axis view plot-x plot-y plot-h (axis-labels y-axis)
                   :thickness thickness
                   :ticks-per-label (axis-ticks-per-label y-axis)))  
    (when coords?
      (draw-coords view plot-x plot-y plot-w plot-h
                   (axis-minimum x-axis) (axis-maximum x-axis)
                   (axis-minimum y-axis) (axis-maximum y-axis)
                   (plot-coords plot)
                   :thickness thickness :radius radius))
    plot))

(defun create-plot (&key (name "Untitled") coords x-axis y-axis external)
  (setf coords (parse-coords coords))
  (unless x-axis
    (let ((min (reduce #'min coords :key #'car :initial-value 0.0))
          (max (reduce #'max coords :key #'car :initial-value 1.0)))
      (setf x-axis (make-axis :minimum min :maximum max
                              :ticks-per-label 5 :divisions 4))
      (setf (axis-labels x-axis) (compute-axis-labels x-axis))))
  (unless y-axis
    (let ((min (reduce #'min coords :key #'cdr :initial-value 0.0))
          (max (reduce #'max coords :key #'cdr :initial-value 1.0)))
      (setf y-axis (make-axis :minimum min :maximum max
                              :ticks-per-label 5 :divisions 4))
      (setf (axis-labels y-axis) (compute-axis-labels y-axis))))
  (make-plot :name name :coords coords :x-axis x-axis :y-axis y-axis
             :external external))

;;;
;;; a plot view displays the axes and points of a plot. its methods
;;; explicitly check for data so the plot editor can simply erase the
;;; display if no plots are loaded, rather then deinstalling the view.
;;; if enabled, the guide view is superimposed on the plot coordinates
;;; and the cursor is accompanied by gray line cross hairs.
;;;

(defclass plot-view (ccl:view)
  ((plot :initarg :plot :initform nil :accessor plot-view-plot)
   (guide? :initform t :accessor plot-view-guide? :initarg :guide?))
  (:default-initargs 
    :view-nick-name ':plot-view
    :view-font '("Geneva" 8 :srcor :plain)))

(defmethod initialize-instance :after ((view plot-view) &rest args)
  (declare (ignore args))
  (when (slot-value view 'plot)
    (set-plot-dimensions view)))

(defmethod ccl:view-cursor ((view plot-view) point)
  ;; guiding uses a cross hairs cursor.
  (declare (ignore point))
  (if (slot-value view 'guide?)
    (cross-hair-cursor); system cross hair cursor
    ccl:*arrow-cursor*))

;(enved (list 0 1 10 .25  25 .1 100 0))
;(enved nil)
      
(defmethod ccl:view-draw-contents ((view plot-view))
  (when (slot-value view 'plot)
    (draw-plot view (plot-view-plot view))))

(defun redraw-plot-view (view &optional (mode t))
  ;; fix this to selectively redraw portions of plot
  (declare (ignore mode))
  (ccl:invalidate-view view t)
  view)

(defmethod ccl:set-view-size :after ((view plot-view) xy &optional y)
  (declare (ignore xy y))
  (let ((plot (slot-value view 'plot)))
    (when plot
      (set-plot-dimensions view)
      (ccl:invalidate-view view t))))

(defmethod set-plot-view-plot (view plot)
  (setf (slot-value view 'plot) plot)
  (when plot (set-plot-dimensions view))
  (ccl:invalidate-view view t))

(defun set-plot-dimensions (view)
  ;; plot is offset to near bottom of view. clim coords were
  ;; aways 0,0 at bottom, as a consequence all the y incrementing
  ;; is negative for mcl!
  (let* ((plot (plot-view-plot view))
         (font (ccl:view-font view))
         (x-axis (plot-x-axis plot))
         (y-axis (plot-y-axis plot))
         (x-labels (axis-labels x-axis))
         (y-labels (axis-labels y-axis))
         (x-length (length x-labels))
         (margin 4)
         plot-x plot-y plot-w plot-h)
    ;; x origin at left + max y label width + tick space
    (setf plot-x 
          (+ margin
             (reduce #'max y-labels
                     :key #'(lambda (s) (ccl:string-width s font)))
             *tick-space*))
    ;; y origin at bottom of view - axis label and tick space
    (setf plot-y 
          (- (view-height view) margin *tick-space* (font-height font)))
    ;; width is right - 1/2 the x centered last label - plot-x
    (setf plot-w
          (- (view-width view)
             (+ margin
                (floor (ccl:string-width
                        (nth (1- x-length) x-labels)
                        font)
                       2)
                plot-x)))
    ;; height is bottom - 1/2 the y centered top label
    (setf plot-h (- plot-y (floor (font-height font) 2))) ;margin
    (setf (plot-position plot) (xy plot-x plot-y))
    (setf (plot-size plot) (xy plot-w plot-h))
    plot))

(defun point-to-coord (point plot)
  (let* ((pos (plot-position plot))
         (size (plot-size plot))
         (x-axis (plot-x-axis plot))
         (y-axis (plot-y-axis plot))
         (x-minimum (axis-minimum x-axis))
         (y-minimum (axis-minimum y-axis))
         (x-range (- (axis-maximum x-axis) x-minimum))
         (y-range (- (axis-maximum y-axis) y-minimum)))
    (coord (+ (* (/ (- (xy-x point) (xy-x pos)) 
                    (xy-x size)) x-range) x-minimum)
           (+ (* (/ (- (xy-y pos) (xy-y point)) 
                    (xy-y size)) y-range) y-minimum))))

(defun coord-to-point (coord plot)
  (let* ((pos (plot-position plot))
         (size (plot-size plot))
         (x-axis (plot-x-axis plot))
         (y-axis (plot-y-axis plot))
         (x-minimum (axis-minimum x-axis))
         (y-minimum (axis-minimum y-axis))
         (x-range (- (axis-maximum x-axis) x-minimum))
         (y-range (- (axis-maximum y-axis) y-minimum)))
    (xy (floor
         (+ (xy-x pos) (* (/ (- (coord-x coord) x-minimum) x-range)
                          (xy-x size))))
        (floor 
         (- (xy-y pos) (* (/ (- (coord-y coord) y-minimum) y-range)
                          (xy-y size)))))))

(defun find-coord (plot point)
  (let ((delta (xy (plot-radius plot) 
                   (- (plot-radius plot))))
        pos1 pos2)
    (setf pos1 (point-to-coord (xy- point delta) plot))
    (setf pos2 (point-to-coord (xy+ point delta) plot))
    (loop for coord in (plot-coords plot)
          until (> (coord-x coord) (coord-x pos2))
          when (and (<= (coord-x pos1) (coord-x coord) (coord-x pos2))
                    (<= (coord-y pos1) (coord-y coord) (coord-y pos2)))
          return coord)))

(defmethod ccl:view-click-event-handler ((view plot-view) where)
  (let ((plot (plot-view-plot view)))
    (when plot
      (ccl:with-focused-view view
        (let ((global (let ((v (enved-aspects-view view)))
                        (if (eq (ccl:view-nick-name v) ':global-aspects)
                          v nil)))
              (coord (find-coord plot where)))
          (cond ((ccl:shift-key-p)
                 (cond (coord
                        (delete-coord view coord)
                        (setf coord nil))
                       (t 
                        (setf coord (point-to-coord where plot))
                        (add-coord view coord))))
                (coord
                 (move-coord view coord)))
          (when global (update-active-coord global coord)))))))

(defun delete-coord (view coord)
  (let* ((plot (plot-view-plot view))
         (pos (plot-position plot))
         (size (plot-size plot))
         (x-axis (plot-x-axis plot))
         (y-axis (plot-y-axis plot))
         (coords (plot-coords plot))
         (loc (position coord (plot-coords plot)))
         (redraw (list coord)))
    (unless (= loc (1- (length coords)))
      (nconc redraw (list (elt coords (1+ loc)))))
    (unless (= loc 0)
      (push (elt coords (1- loc)) redraw))
    (draw-coords view (xy-x pos) (xy-y pos)
                 (xy-x size) (xy-y size)             
                 (axis-minimum x-axis) (axis-maximum x-axis)
                 (axis-minimum y-axis) (axis-maximum y-axis)
                 redraw :thickness (plot-pen plot) 
                 :radius (plot-radius plot) :erase? t)
    (setf (plot-coords plot) (delete coord (plot-coords plot)))
    (when (setf redraw (delete coord redraw))
      (draw-coords view (xy-x pos) (xy-y pos)
                   (xy-x size) (xy-y size)             
                   (axis-minimum x-axis) (axis-maximum x-axis)
                   (axis-minimum y-axis) (axis-maximum y-axis)
                   redraw :thickness (plot-pen plot) 
                   :radius (plot-radius plot)))))

;(enved (list 0 1 10 .25  25 .1 100 0))
;(enved)

(defun add-coord (view coord)
  (flet ((insert (item list)
           (let ((this list) (last nil))
             (loop while 
                   (and this 
                        (not (< (coord-x item) 
                                (coord-x (car this)))))
                   do (setf last this this (cdr this)))
             (setf this (cons item this))
             (if last (rplacd last this) (setf list this))
             list))
         (update-axis? (value axis &aux redo?)
           (when (< (axis-maximum axis) value)
             (setf (axis-maximum axis) value redo? t))
           (when (> (axis-minimum axis) value) 
             (setf (axis-minimum axis) value redo? t))
           (when redo? 
             (setf (axis-labels axis) (compute-axis-labels axis)))
           redo?))
    (let* ((plot (plot-view-plot view))
           (pos (plot-position plot))
           (size (plot-size plot))
           (x-axis (plot-x-axis plot))
           (y-axis (plot-y-axis plot))
           (redraw nil)
           loc)
      (setf (plot-coords plot) (insert coord (plot-coords plot)))
      (setf loc (position coord (plot-coords plot)))
      (unless (= loc (1- (length (plot-coords plot))))
        (push (elt (plot-coords plot) (1+ loc)) redraw))
      (unless (= loc 0)
        (push (elt (plot-coords plot) (1- loc)) redraw))
      (when redraw
        (draw-coords view (xy-x pos) (xy-y pos)
                     (xy-x size) (xy-y size)             
                     (axis-minimum x-axis) (axis-maximum x-axis)
                     (axis-minimum y-axis) (axis-maximum y-axis)
                     redraw :thickness (plot-pen plot) 
                     :radius (plot-radius plot) :erase? t))
      (setf redraw (insert coord redraw))
      (draw-coords view (xy-x pos) (xy-y pos)
                   (xy-x size) (xy-y size)             
                   (axis-minimum x-axis) (axis-maximum x-axis)
                   (axis-minimum y-axis) (axis-maximum y-axis)
                   redraw :thickness (plot-pen plot) 
                   :radius (plot-radius plot)))
    coord))

;(enved (list 0 1 10 .25  25 .1 100 0))
;(enved nil)

(defmethod move-coord (view coord)
  (let* ((plot (plot-view-plot view))
         (pos (plot-position plot))
         (size (plot-size plot))
         (x-axis (plot-x-axis plot))
         (y-axis (plot-y-axis plot))
         (coords (plot-coords plot))
         (newpoint (coord-to-point coord plot))
         (oldpoint newpoint) ; 
         (thick (plot-pen plot))
         (radius (plot-radius plot))
         (guide? (plot-view-guide? view))
         (guideend (xy (+ (xy-x pos) (xy-x size))
                       (- (xy-y pos) (xy-y size))))
         loc lcoord rcoord lpoint rpoint xmin xmax ymin ymax )
    (setf loc (position coord coords))
    (unless (= loc (1- (length coords)))
      (setf rcoord (elt coords (1+ loc)))
      (setf rpoint (coord-to-point rcoord plot)))
    (unless (= loc 0)
      (setf lcoord (elt (plot-coords plot) (1- loc)))
      (setf lpoint (coord-to-point lcoord plot)))
    (setf xmin (xy-x (or lpoint pos)))
    (setf xmax (xy-x (or rpoint (xy+ pos size))))
    (setf ymin (- (xy-y pos) (xy-y size)))
    (setf ymax (xy-y pos))
    (draw-coords view (xy-x pos) (xy-y pos)
                 (xy-x size) (xy-y size)             
                 (axis-minimum x-axis) (axis-maximum x-axis)
                 (axis-minimum y-axis) (axis-maximum y-axis)
                 (list coord) :thickness thick :radius radius
                 :erase? t)
    (flet ((draw-guide (view pos off end)
             (ccl:with-focused-view view
               (ccl:with-pen-saved
                 (ccl:set-pen-mode view ':patxor)
                 (ccl:set-pen-pattern view ccl:*gray-pattern*)
                 (#_MoveTo (xy-x off) (xy-y pos))
                 (#_LineTo (xy-x end) (xy-y pos))
                 (#_MoveTo (xy-x pos) (xy-y off))
                 (#_LineTo (xy-x pos) (xy-y end))))))
      (ccl:without-interrupts
        (loop with on = nil
              while (ccl:mouse-down-p)
              do 
              (setf newpoint (ccl:view-mouse-position view))
              ;; bounds checking
              (unless (eql oldpoint newpoint)
                (if (< (xy-x newpoint) xmin) 
                  (setf newpoint (xy xmin (xy-y newpoint)))
                  (if (> (xy-x newpoint) xmax) 
                    (setf newpoint (xy xmax (xy-y newpoint)))))            
                (if (< (xy-y newpoint) ymin) 
                  (setf newpoint (xy (xy-x newpoint) ymin))
                  (if (> (xy-y newpoint) ymax)
                    (setf newpoint (xy (xy-x newpoint) ymax))))            
                (when guide? 
                  (when on ; dont xor erase oldpoint until newpoint
                    (draw-guide view oldpoint pos guideend)))
                (when lpoint
                  (draw-line view lpoint oldpoint
                             :thickness thick :erase? t)
                  (draw-line view lpoint newpoint :thickness thick))
                (when rpoint
                  (draw-line view oldpoint rpoint
                             :thickness thick :erase? t)
                  (draw-line view newpoint rpoint :thickness thick))
                (when guide?
                  (draw-guide view newpoint pos guideend)
                  (setf on t))
                ;; moving can clobber the axis display
                (draw-x-axis view (xy-x pos) (xy-y pos) (xy-x size) nil
                             :thickness thick
                             :ticks-per-label
                             (axis-ticks-per-label x-axis))
                (draw-y-axis view (xy-x pos) (xy-y pos) (xy-y size) nil
                             :thickness thick
                             :ticks-per-label
                             (axis-ticks-per-label y-axis)))
              (setf oldpoint newpoint) )
        (when guide? (draw-guide view newpoint pos guideend))))
    (let ((c (point-to-coord newpoint plot)))
      (setf (coord-x coord) (coord-x c))
      (setf (coord-y coord) (coord-y c)))
    (draw-coords view (xy-x pos) (xy-y pos)
                 (xy-x size) (xy-y size)             
                 (axis-minimum x-axis) (axis-maximum x-axis)
                 (axis-minimum y-axis) (axis-maximum y-axis)
                 coords ;(list coord)
                 :thickness thick :radius radius)
    coord))


;;;
;;; aspects are views for editing various plot characteristics. 
;;; when one is selected from a pop up menu it is installed as
;;; subview of the main contol view.  The current aspects are:
;;; global-aspects - plot name, line and circle thickness
;;; axis-aspects   - axis labels, min and max, ticks and divisions
;;; coord-aspects  -  X and Y typein buffer, cursor guide
;;; output-aspects - exporting format (string, list, etc.)
;;;

(defclass enved-aspects (ccl:view bgc-mixin)
  ((data :initarg :data :initform nil :accessor aspects-data))
  (:default-initargs 
    :view-nick-name ':global-aspects
    :view-font *capella-font* ))

;;; aspects are subviews of the control view and pass
;;; enable/disable commands down to their subitems.


(defmethod control-enable ((view enved-aspects))
  (map nil #'control-enable (ccl:view-subviews view)))

(defmethod control-disable ((view enved-aspects))
  (map nil #'control-disable (ccl:view-subviews view)))

(defmethod update-aspects ((view enved-aspects) data)
  ;; default does nothing
  data)

(defmethod update-aspects :before ((view enved-aspects) data)
  (setf (slot-value view 'data) data))

;;;
;;; Global aspects
;;; 

(defclass global-aspects (enved-aspects)
  ((coord :initform nil :accessor global-aspects-active-coord))
  (:default-initargs 
    :view-nick-name ':global-aspects))

(defmethod initialize-instance :after ((view global-aspects) &rest args)
  (declare (ignore args))
  (let ((s0 (string-extent "Name:" *capella-font* (xy 3 1)))
        (s1 (string-extent "Untitled-100" *capella-font* (xy 0 -1)))
        (s2 (string-extent "Line Size:" *capella-font* (xy 3 1)))
        (s3 (string-extent "000" *capella-font* (xy 0 -1)))
        (s4 (string-extent "Point Size:" *capella-font* (xy 3 1)))
        (s5 (string-extent "X:" *capella-font* (xy 3 1)))
        (s6 (string-extent "1234.567" *capella-font* (xy 0 -1)))
        (s7 (string-extent "Mouse Guide" *capella-font* (xy 20 1)))
        (w 4)
        x y naml namb linl linb cirl cirb xlab xbuf ylab ybuf guide)
    (setf x w y 4)
    (setf naml (make-instance 'label :title "Name:"
                              :view-position (xy x y) :view-size s0))
    (setf x (+ x (view-width naml) w))
    (setf namb (make-instance 'textbuf :view-nick-name ':name
                              :view-position (xy x y) :view-size s1))    
    (setf x (+ x (view-width namb) (* w 2)))
    (setf linl (make-instance 'label :title "Line Size:" 
                              :view-position (xy x y) :view-size s2))
    (setf x (+ x (view-width linl) w))
    (setf linb (make-instance 'textbuf :view-nick-name ':line
                              :view-position (xy x y) :view-size s3))
    (setf x (+ x (view-width linb) (* w 2)))
    (setf cirl (make-instance 'label :title "Point SIze:" 
                              :view-position (xy x y) :view-size s4))
    (setf x (+ x (view-width cirl) w))
    (setf cirb (make-instance 'textbuf :view-nick-name ':coord
                              :view-position (xy x y) :view-size s3))

    (setf y (+ (floor (view-height view) 2) w))
    (setf x w)
    (setf xlab (make-instance 'label :title "X:"
                              :view-position (xy x y) :view-size s5))
    (setf x (+ x (view-width xlab) w))
    (setf xbuf (make-instance 'textbuf :dialog-item-text ""
                              :view-position (xy x y) :view-size s6
                              :view-nick-name ':xbuf))
    (setf x (+ x (view-width xbuf) (* w 2)))
    (setf ylab (make-instance 'label :title "Y:"
                              :view-position (xy x y) :view-size s5))
    (setf x (+ x (view-width ylab) w))
    (setf ybuf (make-instance 'textbuf :dialog-item-text ""
                              :view-position (xy x y) :view-size s6 
                              :view-nick-name ':ybuf))
    (setf x (+ x (view-width ybuf) w w w w))
    (setf guide (make-instance 'check-box
                  :title "Mouse guide" :selected? t 
                  :view-nick-name ':guide
                  :view-position (xy x y) :view-size s7
                  :action 
                  #'(lambda (v)
                      (setf (plot-view-guide? (enved-plot-view v))
                            (toggle-selected? v)))))

    (ccl:add-subviews 
     view (make-instance 'invisible-default-button
            :action #'global-aspects-action)
     naml namb linl linb cirl cirb xlab xbuf ylab ybuf guide)))

;(enved (list 0 1 10 .25  25 .1 100 0))
;(enved)

(defmethod update-active-coord ((view global-aspects) coord)
  (setf (slot-value view 'coord) coord)
  (let ((xbuf (find-subview view ':xbuf))
        (ybuf (find-subview view ':ybuf)))
    (ccl:set-dialog-item-text
     xbuf (if (null coord) "" (format nil "~,2F" (coord-x coord))))
    (ccl:set-dialog-item-text 
     ybuf (if (null coord) "" (format nil "~,2F" (coord-y coord))))))

(defun global-aspects-action (view)
  (setf view (ccl:view-container view))  ; was inivisible button.
  (let ((buf (ccl:current-key-handler (ccl:view-window view)))
        (plot (enved-plot view))
        (redraw? nil)
        value)
    (case (and buf (ccl:view-nick-name buf))
      (:name
       (unless (or (string= (setf value (textbuf-text buf)) "")
                   (string-equal value (or (plot-name plot) "")))
         (rename-plot view plot value)))
      (:line
       (when (and (setf value (textbuf-integer buf))
                  (/= value (plot-pen plot)))
         (setf (plot-pen plot) value)
         (setf redraw? ':coords)))
      (:coord
       (when (and (setf value (textbuf-integer buf))
                  (/= value (plot-radius plot)))
         (setf (plot-radius plot) value)
         (setf redraw? ':coords)))
      ((:xbuf :ybuf)
       (let ((coord (global-aspects-active-coord view)))
         (when (and coord
                    (numberp (setf value (textbuf-form buf))))
           (if (eq (ccl:view-nick-name buf) ':xbuf)
             (setf (coord-x coord) value)
             (setf (coord-y coord) value))
           (setf redraw? ':coords)))))
    (when redraw? 
      (redraw-plot-view (enved-plot-view view) redraw?))))

(defmethod update-aspects ((view global-aspects) plot)
  (ccl:set-dialog-item-text (find-subview view ':name)
                            (or (plot-name plot) ""))
  (ccl:set-dialog-item-text (find-subview view ':line)
                            (prin1-to-string (plot-pen plot)))
  (ccl:set-dialog-item-text (find-subview view ':coord)
                            (prin1-to-string (plot-radius plot)))
  (update-active-coord view nil))

;;;
;;; Axis aspects.
;;;

(defclass axis-aspects (enved-aspects)
  ())

(defmethod initialize-instance :after ((view axis-aspects) &rest args)
  (declare (ignore args))
  (let ((s0 (string-extent "Label:" *capella-font* (xy 3 0)))
        (s1 (string-extent "Untitled" *capella-font* (xy 0 -2)))
        (s2 (string-extent "Min:" *capella-font* (xy 3 0)))
        (s3 (string-extent "1234.567" *capella-font* (xy 0 -2)))
        (s4 (string-extent "123" *capella-font* (xy 0 -2)))
        (s5 (string-extent "Divisions:" *capella-font* (xy 3 0)))
        (s6 (string-extent "Ticks:" *capella-font* (xy 3 0)))
        (s7 (string-extent "Decimals:" *capella-font* (xy 3 0)))
        labl lab minl min maxl max divl div ticl tic decl dec)
    (setf labl (make-instance 'label :title "Label:"
                              :view-nick-name ':labl :view-size s0))
    (setf lab (make-instance 'textbuf :dialog-item-text ""
                             :dialog-item-enabled-p nil
                             :view-nick-name ':lab :view-size s1))
    (setf minl (make-instance 'label :title "Min:"
                              :view-nick-name ':minl :view-size s2))
    (setf min (make-instance 'textbuf :dialog-item-text "1234.567"
                             :view-nick-name ':min :view-size s3))
    (setf maxl (make-instance 'label :title "Max:"
                              :view-nick-name ':maxl :view-size s2))
    (setf max (make-instance 'textbuf :dialog-item-text "1234.567"
                             :view-nick-name ':max :view-size s3))
    (setf divl (make-instance 'label :title "Divisions:"
                              :view-nick-name ':divl :view-size s5))
    (setf div (make-instance 'textbuf :dialog-item-text ""
                             :view-nick-name ':div :view-size s4))
    (setf ticl (make-instance 'label :title "Ticks:"
                              :view-nick-name ':ticl :view-size s6))
    (setf tic (make-instance 'textbuf :dialog-item-text ""
                             :view-nick-name ':tic :view-size s4))
    (setf decl (make-instance 'label :title "Decimals:"
                              :view-nick-name ':decl
                              :view-size s7))
    (setf dec (make-instance 'textbuf :dialog-item-text ""
                             :view-nick-name ':dec :view-size s4))
    (ccl:add-subviews 
     view
     (make-instance 'invisible-default-button
       :action #'axis-aspects-action)
     labl lab minl min maxl max divl div ticl tic decl dec)
    (position-axis-aspects view)))

(defun position-axis-aspects (view)
  (let ((size (ccl:view-size view))
        (m 0)
        (s 8)
        (labl (find-subview view ':labl))
        (lab (find-subview view ':lab))
        (minl (find-subview view ':minl))
        (min (find-subview view ':min))
        (maxl (find-subview view ':maxl))
        (max (find-subview view ':max))
        (divl (find-subview view ':divl))
        (div (find-subview view ':div))
        (ticl (find-subview view ':ticl))
        (tic (find-subview view ':tic))
        (decl (find-subview view ':decl))
        (dec (find-subview view ':dec))
        x y)
    (setf x m y 3)
    (ccl:set-view-position labl (xy x y))
    (setf x (+ x (view-width labl) s))
    (ccl:set-view-position lab (xy x y))
    (setf x (floor (* (xy-x size) 1/3)))
    (ccl:set-view-position minl (xy x y))
    (setf x (+ x (view-width minl) s))
    (ccl:set-view-position min (xy x y))
    (setf x (floor (* (xy-x size) 2/3)))
    (ccl:set-view-position maxl (xy x y))
    (setf x (+ x (view-width maxl) s))
    (ccl:set-view-position max (xy x y))
    ;; second line
    ;(setf x m y (+ y (view-height min) s))
    (setf x m y (+ (floor (view-height view) 2) 3))
    (ccl:set-view-position divl (xy x y))
    (setf x (+ x (view-width divl) s))
    (ccl:set-view-position div (xy x y))
    (setf x (floor (* (xy-x size) 1/3)))
    (ccl:set-view-position ticl (xy x y))
    (setf x (+ x (view-width ticl) s))
    (ccl:set-view-position tic (xy x y))
    (setf x (floor (* (xy-x size) 2/3)))
    (ccl:set-view-position decl (xy x y))
    (setf x (+ x (view-width decl) s))
    (ccl:set-view-position dec (xy x y))))

(defmethod update-aspects ((view axis-aspects) plot)
  (let* ((axis (ecase (ccl:view-nick-name view)
                 (:x-axis (plot-x-axis plot))
                 (:y-axis (plot-y-axis plot))))
         (labels (axis-labels axis)))
    (ccl:set-dialog-item-text 
     (find-subview view ':min) (if plot (pop labels) ""))
    (ccl:set-dialog-item-text
     (find-subview view ':max) (if plot (first (last labels)) ""))
    (ccl:set-dialog-item-text
     (find-subview view ':div) 
     (if plot (prin1-to-string (axis-divisions axis)) ""))
    (ccl:set-dialog-item-text 
     (find-subview view ':tic) 
     (if plot (prin1-to-string (axis-ticks-per-label axis)) ""))
    (ccl:set-dialog-item-text 
     (find-subview view ':dec)
     (if plot (prin1-to-string (axis-decimals axis)) ""))))

(defun axis-aspects-action (view)
  (setf view (ccl:view-container view))  ; was inivisible button.
  (let ((min (find-subview view ':min))
        (max (find-subview view ':max))
        (div (find-subview view ':div))
        (tic (find-subview view ':tic))
        (dec (find-subview view ':dec))
        (plot (aspects-data view))
        (redraw? nil)
        (side (ccl:view-nick-name view))
        axis value)
    (setf axis
          (ecase side
            (:x-axis (plot-x-axis plot))
            (:y-axis (plot-y-axis plot))))
    (when (and (setf value (textbuf-integer min))
               (/= value (axis-minimum axis)))
      (setf (axis-minimum axis) value)
      (setf redraw? side))
    (when (and (numberp (setf value (textbuf-form max)))
               (> value 0)
               (/= value (axis-maximum axis)))
      (setf (axis-maximum axis) value)
      (setf redraw? side))
    (when (and (setf value (textbuf-integer div)) 
               (> value 0)
               (/= value (axis-divisions axis)))
      (setf (axis-divisions axis) value)
      (setf redraw? side))
    (when (and (setf value (textbuf-integer tic))
               (> value 0)
               (/= value (axis-ticks-per-label axis)))
      (setf (axis-ticks-per-label axis) value)
      (setf redraw? side))
    (when (and (setf value (textbuf-integer dec))
               (>= value 0)
               (/= value (axis-decimals axis)))
      (setf (axis-decimals axis) value)
      (setf redraw? side))
    (when redraw?
      (setf (axis-labels axis) (compute-axis-labels axis))
      (set-plot-dimensions (enved-plot-view view))
      (redraw-plot-view (enved-plot-view view) redraw?))))    

;;;
;;; Output aspects.
;;;

(defclass output-aspects (enved-aspects)
  ((format :initform (list :list :float)
           :accessor output-aspects-format))
  (:default-initargs 
    :view-nick-name ':output-aspects))

(defmethod initialize-instance :after ((view output-aspects) &rest args)
  (declare (ignore args))
  (let ((size (ccl:view-size view))
        (s0 (string-extent "Integer" *capella-font* (xy 20 0)))
        (m 4) x y menu float int)
    (setf x m y m)
    (setf menu
          (make-instance 'pop-up-menu
            :view-position (xy x y)
            :menu-items
            (list 
             (make-instance 'ccl:menu-item
               :menu-item-title "(x y ...)"
               :menu-item-action
               #'(lambda () 
                   (setf (car (output-aspects-format view)) ':list)))
             (make-instance 'ccl:menu-item
               :menu-item-title "x y ..."
               :menu-item-action
               #'(lambda () 
                   (setf (car (output-aspects-format view)) ':string)))
             (make-instance 'ccl:menu-item
               :menu-item-title  "x,y ..."
               :menu-item-action
               #'(lambda () 
                   (setf (car (output-aspects-format view)) ':comma))))))
    (setf y (+ (floor (xy-y size) 2) m))
    (setf float (make-instance 'radio-button
                  :title "Float" :selected? t
                  :action
                  #'(lambda (b) b
                     (setf (cadr (output-aspects-format view))
                           ':float))
                  :view-position (xy x y) :view-size s0))
    (setf x (+ x (view-width float) m))
    (setf int (make-instance 'radio-button
                :title "Integer" :selected? nil
                :action
                #'(lambda (b) b
                   (setf (cadr (output-aspects-format view))
                         ':integer))
                :view-position (xy x y) :view-size s0))
    (ccl:add-subviews view menu float int)))
  
;;;
;;; main control view holds top-level buttons and pop up menus
;;; to install plots and aspects.
;;;

(defclass enved-control (gray-3d-framed-view)
  ((plots :initform nil :initarg :plots :accessor enved-plots)
   (aspects :initform nil  :initarg :aspects :accessor enved-aspects))
  (:default-initargs 
    :view-nick-name ':enved-control
    :view-font *capella-font*))

;(enved (list 0 1 10 .25  25 .1 100 0))
;(enved)

(defmethod initialize-instance :after ((view enved-control) &rest args)
  (declare (ignore args))
  (let* ((s0 (string-extent "Untitled-100" *capella-font* (xy 20 5)))
         (s1 (string-extent "Output   " *capella-font* (xy 20 5)))
         (s2 (string-extent "Delete " *capella-font* (xy 2 2)))
         (m 8)
         (size (ccl:view-size view))
         (plots (slot-value view 'plots))
         x y menu1 menu2 add del dum axis global output)
    (setf x m y m)
    (setf menu1 (make-instance 'pop-up-menu
                  :view-nick-name ':plots
                  :menu-items 
                  (loop for p in plots
                        collect 
                        (let ((p p))
                          (make-instance 'ccl:menu-item
                            :menu-item-title (plot-name p)
                            :menu-item-action
                            #'(lambda () (switch-plots view p)))))
                  :view-position (xy x y) :view-size s0))
    ;; right justify buttons
    (setf x (- (xy-x size) 8 (xy-x s2)))
    (setf dum (make-instance 'button :title "Dump"
                             :action
                             #'(lambda (b)
                                 (dump-plot (enved-control-view b)))
                             :view-nick-name ':dump
                             :view-position (xy x y) :view-size s2))
    (setf x (- x (xy-x s2) 4))
    (setf del (make-instance 'button :title "Delete"
                             :action
                             #'(lambda (b)
                                 (delete-plot (enved-control-view b)))
                             :view-nick-name ':delete
                             :view-position (xy x y) :view-size s2))
    (setf x (- x (xy-x s2) 4))
    (setf add (make-instance 'button :title "New"
                             :action
                             #'(lambda (b)
                                 (add-plot (enved-control-view b)))
                             :view-nick-name ':new
                             :view-position (xy x y) :view-size s2))
    (setf x (- (floor (xy-x size) 2) (floor (xy-x s1) 2)))
    ;(setf y (+ (floor (* (xy-y size) 1/4)) 5))
    ;(setf y (+ (view-top menu1) (floor (view-height menu1) 2)))
    (setf y (+ (view-bottom menu1) 0))
    (setf menu2 
          (make-instance 'pop-up-menu
             :view-position (xy x y) :view-size s1
            :menu-items 
            (list
             (make-instance 'ccl:menu-item
               :menu-item-title "Global"
               :menu-item-action
               #'(lambda () (switch-aspects view ':global-aspects)))
             (make-instance 'ccl:menu-item
               :menu-item-title "X Axis"
               :menu-item-action
               #'(lambda (&aux (n ':x-axis))
                   (ccl:set-view-nick-name axis n)
                   (switch-aspects view n)))
             (make-instance 'ccl:menu-item
               :menu-item-title "Y Axis"
               :menu-item-action
               #'(lambda (&aux (n ':y-axis))
                   (ccl:set-view-nick-name axis n)
                   (switch-aspects view n)))
             (make-instance 'ccl:menu-item
               :menu-item-title "Output"
               :menu-item-action
               #'(lambda () (switch-aspects view ':output-aspects))))
          ;  :view-position (xy x y) :view-size s0
            ))
    (setf y (floor (* (xy-y size) 1/2)))
    ;; create the aspect editing views and cache. make global be the
    ;; active aspect.
    (let ((p (xy 5 y))
          (s (xy (- (xy-x size ) 16) (- (floor (xy-y size) 2) 10))))
      (setf axis (make-instance 'axis-aspects 
                   :view-position p :view-size s))
      (setf global (make-instance 'global-aspects
                     :view-position p :view-size s))
      (setf output (make-instance 'output-aspects
                     :view-position p :view-size s))
      ;(when plot (update-aspects global plot))
      (setf (enved-aspects view)
            (list global axis output)))
    (ccl:add-subviews view menu1 add del dum menu2 global)))

(defmethod ccl:set-view-size :after ((view enved-control) xy &optional y)
  (declare (ignore y))
  (dolist (a (enved-aspects view))
    (ccl:set-view-size a (xy (- (xy-x xy) 16) (view-height a)))))

(defun switch-aspects (control aspect)
  (setf aspect (find aspect (enved-aspects control)
                     :key #'ccl:view-nick-name))
  (let ((plot (enved-plot control)))
    (unless (find aspect (ccl:view-subviews control))
      (dolist (v (enved-aspects control))
        (when (eq (ccl:view-container v) control)
          (ccl:remove-subviews control v))))
    (ccl:add-subviews control aspect)
    (update-aspects aspect plot)))

(defun switch-plots (control plot)
  (let ((aspects (enved-aspects-view control))
        (display (enved-plot-view control)))
    (unless (eq plot (plot-view-plot display))
      (update-aspects aspects plot)
      (set-plot-view-plot display plot))))

(defun add-plot (control &optional plot coords external)
  (let* ((plots (enved-plots control))
         (menu (find-subview control ':plots))
         item)
    (unless plot
      (setf plot (create-plot :name
                              (if (null plots) "Untitled"
                                  (format nil "Untitled-~A" 
                                          (1+ (length plots))))
                              :coords coords
                              :external external)))
    (setf (enved-plots control) (nconc plots (list plot)))
    (setf item (make-instance 'ccl:menu-item
                 :menu-item-title (plot-name plot)
                 :menu-item-action 
                 #'(lambda () (switch-plots control plot))))
    (ccl:add-menu-items menu item)
    ;; switch to new plot by selecting it in the pop-up
    (pop-up-menu-select menu item)
    (enved-update control)))

(defun delete-plot (control &optional plot)
  (unless plot (setf plot (enved-plot control)))
  (let* ((plots (enved-plots control))
         (menu (find-subview control ':plots))
         (items (ccl:menu-items menu))
         (max (1- (length items)))
         (pos (position plot plots)))
    (setf plots (delete plot plots))
    (setf (enved-plots control) plots)
    ;; assumes mcl adds items to the end of the list!
    (ccl:remove-menu-items menu (elt items pos))
    (if plots
      (progn
        (if (= pos max) (decf pos))
        (pop-up-menu-select menu (elt items pos)))
      (switch-plots control nil))
    (enved-update control)))

(defun enved-update (view)
  (setf view (enved-control-view view))
  (let ((subs (ccl:subviews view ))
        (help (ccl:find-named-sibling view ':help)))
    (if (= (length (enved-plots view)) 0)
      (progn
        (dolist (v subs)
          (unless (eq (ccl:view-nick-name v) ':new)
            (control-disable v)))
        (control-disable help))
      (progn
        (dolist (v subs)
          (unless (eq (ccl:view-nick-name v) ':new)
            (control-enable v)))
        (control-enable help)))))
        
(defun rename-plot (view plot name)
  (setf view (enved-control-view view))
  (let ((plots (enved-plots view))
        (menu (find-subview view ':plots)))
    (setf (plot-name plot) name)
    (let ((item (elt (ccl:menu-items menu) 
                     (position plot plots))))
      (ccl:set-menu-item-title item name)
      (ccl:invalidate-view menu))))

(defun dump-plot (control &optional pasteboard?)
  (let* ((p (enved-plot control))
         (c (unparse-coords (plot-coords p)))
         (f (enved-output-format control))
         (s ; what a mess! 
          (if (eq (second f) ':float)
            (let ((d (max (axis-decimals (plot-x-axis p))
                          (axis-decimals (plot-y-axis p)))))
              (ecase (first f)
                (:list (format nil "(~~,~Df~~{ ~~,~Df~~})" d d))
                (:string  (format nil "~~,~Df~~{ ~~,~Df~~}" d d))
                (:comma (format nil "~~,~Df~~{,~~,~Df~~}" d d))))
            (ecase (first f)
              (:list "(~D~{ ~D~})")
              (:string "~D~{ ~D~}")
              (:comma "~,3f~{,~,3f~}")))))
    (if pasteboard?
      (ccl:put-scrap ':text (format nil s (pop c) c))
      (let ((w (make-instance 'ccl:fred-window
                 :window-show nil :view-size #@(200 100)
                 :window-title
                 (format nil "~A Envelope" (plot-name p)))))
        (format w s (pop c) c)
        (ccl:fred-update w)
        (ccl:window-set-not-modified #-ccl-3 w #+ccl-3 (ccl:fred-item w))
        (ccl:window-select w)))))

;;;
;;; (enved (list 1 1 10 .25  25 .1 100 0))
;;;

(defparameter *default-enved-size* (xy 300 250))

(defclass enved (capella-window)
  ()
  (:default-initargs 
    :view-size *default-enved-size*
    :window-title "Envelope Editor" ))
  
(defmethod initialize-instance :after ((view enved) &rest args)
  (declare (ignore args))
  (set-window-grow view *default-enved-size*))

(defmethod ccl:window-size-parts ((view enved))
  (let ((size (ccl:view-size view))
        (help (find-subview view ':help))
        (plot (find-subview view ':plot-view))
        (main (find-subview view ':enved-control)))
    (when main
      (ccl:set-view-size 
       main (xy (view-width view) (view-height main))))
    (when help 
      (ccl:set-view-position 
       help (xy (- (floor (xy-x size) 2)
                   (floor (xy-x (view-width help)) 2))
                (- (xy-y size) 14))))
    (when plot
      (ccl:set-view-size 
       plot (xy (xy-x size) (- (view-top help) (view-bottom main)))))))

(defmethod ccl:window-can-do-operation ((window enved)
                                        op &optional item)
  (declare (ignore item))
  (or (eq op 'ccl:copy) (eq op 'ccl:paste)))

(defmethod ccl:copy ((window enved))
  (dump-plot (enved-control-view window) t))

(defmethod ccl:paste ((window enved))
  (let ((form (ccl:get-scrap ':lisp)))
    (if (listp form)
      (let ((ex form))
        (case (first form)
          (interpl (setf form (cddr form)))
          ((list interpolation) (setf form (cdr form)))
          ((function-value interp lookup) (setf form (third form))))
        (when (eq (first form) 'quote)
          (setf form (second form)))
        (if (and (consp form)
                 (loop for z = most-negative-single-float then x
                       for (x y) on form by #'cddr
                       always (and (numberp x) (numberp y) (<= z x))))
          (add-plot (enved-control-view window) nil form ex)
          (warn-user "Clipboard doesn't contain envelope data: ~S"
                     form)))
      (warn-user "Clipboard doesn't contain envelope data: ~S" form))))

;(enved (list 0 1 10 .25  25 .1 100 0))
;(enved nil)
      
(defun enved (&optional envelope)
  (let ((help "Points: Drag to move, Shift-Click to Add or Delete.")
        (rise (font-height *capella-font*))
        plot main w)
    (setf main (make-instance 'enved-control
                 :view-position (xy 0 0) 
                 :view-size (xy (xy-x *default-enved-size*)
                                (* (+ 10 rise 5 rise 10) 2))))
    (setf plot (make-instance 'plot-view
                 :view-position (xy 0 (view-bottom main))))
    (setf help (make-instance 
                 'label :title help
                 :view-nick-name ':help
                 :view-size 
                 (string-extent help *capella-font* (xy 3 0))))
    (setf w (make-instance 'enved
              :view-subviews (list main plot help)))
    (when envelope 
      (add-plot main (create-plot :name "Untitled" :coords envelope)))
    (enved-update w)
    (ccl:window-select w)
    w))

(defun enved-plot (v)
  ;; current plot
  (plot-view-plot (enved-plot-view v)))

(defun enved-plot-list (v)
  ;; list of all plots
  (enved-plots (enved-control-view v)))

(defun enved-plot-view (v)
  ;; plot view
  (find-subview (ccl:view-window v) ':plot-view))

(defun enved-control-view (v)
  ;; control view
  (find-subview (ccl:view-window v) ':enved-control))

(defun enved-output-format (view)
  (output-aspects-format
   (find ':output-aspects
         (enved-aspects (enved-control-view view))
         :key #'ccl:view-nick-name)))

(defun enved-aspects-view (view)
  (first (ccl:subviews (enved-control-view view) 'enved-aspects)))

