;;;-*- Mode: Lisp; Package: STELLA -*-
;;;
;;; seq worksheet
;;;
;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)


;-------------------------------------------------------------------------------
;
; seq view
; 

(defclass sequence-pane (gray-3d-framed-view)
  ((active :initform nil :initarg :sequencer-active 
           :accessor sequencer-active)))


(defun make-sequence-pane () 
  ;; has 3 labels with associated typeins: start, repeat, pause
  (let ((win (make-instance 'ccl:window         ; dummy win to add the pane to
               :window-type :document
               :window-show nil))
        (sequence-pane (make-instance 'sequence-pane
                         :view-nick-name 'sequence-pane
                         :view-subviews 
                         (list (make-instance 'label
                                 :title "Start:"
                                 :title-justification :right
                                 :view-nick-name 'start-label)
                               (make-instance 'label
                                 :title "Repeat:"
                                 :title-justification :right
                                 :view-nick-name 'repeat-label)
                               (make-instance 'label
                                 :title "Pause:"
                                 :title-justification :right
                                 :view-nick-name 'pause-label)
                               (make-instance 'textbuf
                                 :title "888"    ; dummy text
                                 :view-nick-name 'start-typein)
                               (make-instance 'textbuf
                                 :title ""
                                 :view-nick-name 'repeat-typein)
                               (make-instance 'textbuf
                                 :title  ""
                                 :view-nick-name 'pause-typein)))))
    (declare (dynamic-extent win))
    ;; this recursively fixes all sizes 
    (ccl:add-subviews win sequence-pane)
    (ccl:remove-subviews win sequence-pane)
    (ccl:window-close win)
    (let* ((titem (ccl:view-named 'start-typein sequence-pane))
           (tsize (item-size titem))
           (lsize (max (item-size (ccl:view-named 'start-label sequence-pane))
                       (item-size (ccl:view-named 'repeat-label sequence-pane))
                       (item-size (ccl:view-named 'pause-label sequence-pane)))))
      (ccl:set-dialog-item-text titem "")
      ;; make sure all labels and typeins are of equal size
      (map nil #'(lambda (x)
                   (size-item x (if (typep x 'label) lsize tsize)))
           (ccl:view-subviews sequence-pane)))
    sequence-pane))
  
(defun distribute (int divisor)
  (values-list (loop for i from 1 to divisor
                     with fac = (/ int divisor)
                     with old = 0
                     collect (- (- old (setf old (round (* i fac))))))))


(defun resize-sequence-pane (pane &optional h v)
  ;; guidelines: label-left, typein-left (centered);
  ;;             start-top, repeat-top, pause-top. 
  ;; variables:  v-separation, h-separation, label-width, typein-width
  (multiple-value-setq (h v) (canonicalize-point-args :values h v))
  (let* ((start-label (ccl:view-named 'start-label pane))
         (start-typein (ccl:view-named 'start-typein pane))
         (repeat-label (ccl:view-named 'repeat-label pane))
         (repeat-typein (ccl:view-named 'repeat-typein pane))
         (pause-label (ccl:view-named 'pause-label pane))
         (pause-typein (ccl:view-named 'pause-typein pane))
         (label-size (max (item-size start-label)
                          (item-size repeat-label)
                          (item-size pause-label)))
         (typein-size (max (item-size start-typein)
                           (item-size repeat-typein)
                           (item-size pause-typein)))
         (label-width (xy-x label-size))
         (typein-width (xy-x typein-size))
         (row-height (max (xy-y label-size) (xy-y typein-size)))
         (h-separation (floor (/ row-height 4)))
         (v-separation h-separation)
         (resized? nil)
         (margin 11)
         (min-width (+ (* 2 margin) 
                       (* 2 (max label-width typein-width))
                       h-separation))
         ;; horizontally expandable measurements are: left- and right-side space
         (left-space (max (- typein-width label-width) 0))
         (right-space (max (- label-width typein-width) 0))
         (width (if h
                  (let ((diff (- h min-width)))
                    (if (>= diff 0)
                      (progn 
                        (multiple-value-bind (l r) (distribute diff 2)
                          (incf left-space l)
                          (incf right-space r))
                        (setf resized? t)
                        (+ min-width diff))
                      min-width))
                  min-width))
         (min-height (+ (* 2 margin) 
                        (* 3 row-height)
                        (* 2 v-separation)))
         ;; bottom-space is the only vertical expandable measurement 
         (bottom-space 0)
         (height (if h 
                   (let ((diff (- v min-height)))
                     (if (>= diff 0)
                       (progn 
                         (setf bottom-space diff)
                         (setf resized? t)
                         (+ min-height diff))
                       min-height))
                   min-height))
         ;; now the guidelines
         (typein-left (+ margin left-space label-width h-separation))
         (label-left (+ margin left-space))
         (start-top margin)
         (repeat-top (+ start-top row-height v-separation))
         (pause-top (+ repeat-top row-height v-separation)))
    (declare (ccl:ignore-if-unused left-space right-space bottom-space))
    (when (/= (+ margin left-space label-width h-separation typein-width 
                 right-space margin)
              width)
      (warn "sum of horiz. measurements /= width! margin: ~a lspace: ~a lwidth: ~a ~
             h-sep: ~a twidth: ~a rspace: ~a => width: ~a" 
            margin left-space label-width h-separation typein-width right-space width))
    (when (/= (+ margin (* 3 row-height) (* 2 v-separation) bottom-space margin)
              height)
      (warn "sum of vert. measurements /= height! margin: ~a row: ~a v-sep: ~a ~
             bspace: ~a => height: ~a" 
            margin row-height v-separation bottom-space height))
    ;; position items; labels have to move 3 pixels down to compensate for the 
    ;; typein frames
    (position-item start-label label-left (+ start-top 3))
    (position-item repeat-label label-left (+ repeat-top 3))
    (position-item pause-label label-left (+ pause-top 3))
    (position-item start-typein typein-left start-top)
    (position-item repeat-typein typein-left repeat-top)
    (position-item pause-typein typein-left pause-top)
    
    ;; resize pane
    (size-item pane width height)
    (values (item-size pane) resized?)))




(defmethod enable-worksheet ((view sequence-pane))
  (map nil #'control-enable (ccl:subviews view 'textbuf)))

(defmethod disable-worksheet ((view sequence-pane))
  (map nil #'control-disable (ccl:subviews view 'textbuf)))


;-------------------------------------------------------------------------------
;
; sequence worksheet
;
;

(defclass sequence-worksheet (worksheet)
  ()
  (:default-initargs 
    :window-show nil
    :window-title "Sequence"
    :color-p t))


;(defmethod initialize-instance :after ((view sequence-worksheet) &rest args)
;  (declare (ignore args)))


(defmethod enable-worksheet ((view sequence-worksheet))
  (enable-worksheet (ccl:view-named 'sequence-pane view)))

(defmethod disable-worksheet ((view sequence-worksheet))
  (disable-worksheet (ccl:view-named 'sequence-pane view)))

(defmethod initialize-worksheet ((view sequence-worksheet))
  (update-stream-menu (ccl:view-named 'output-pane view)) 
  (if *browser*
    (let ((refs (if (and (listing-browser? *browser*)
                         (any-selection? *browser*))
                  (selections-to-reference (browser-listing *browser*))
                  (list (object-view-object *browser*)))))
      (setf (slot-value (ccl:view-named 'sequence-pane view) 'active) refs)
      (enable-worksheet view)
      ;; don't mess with the title
      ;(set-worksheet-title view "Sequence" *browser*)
      )
    (disable-worksheet view)))


(defmethod initialize-instance :after ((window sequence-worksheet) &rest ignore)
  (declare (ignore ignore))
  (let* ((output-pane (make-output-stream-pane))
         (sequence-pane (make-sequence-pane))
         osize ssize w)
    (ccl:add-subviews window output-pane sequence-pane 
                      (make-instance 'invisible-default-button
                        :action #'worksheet-action))
    (setf osize (resize-output-pane output-pane))
    (setf ssize (resize-sequence-pane sequence-pane))
    (setf w (max (xy-x osize) (xy-x ssize)))
    ;; adjust widths of sequence- and output-pane
    (when (/= (xy-x osize) (xy-y ssize))
      (if (< (xy-x osize) (xy-y ssize))
        (resize-output-pane output-pane w (xy-y osize))
        (resize-sequence-pane sequence-pane w (xy-y ssize))))
    ;; s.th. other than document??
    ;(setf (slot-value window 'ccl::grow-icon-p) nil)
    (size-item window (xy-x osize) (xy+ (xy-y osize) (xy-y ssize)))
    (position-item sequence-pane 0 (xy-y osize))
    (ccl:window-show window)))


(defmethod worksheet-action ((view sequence-worksheet))
  ;(ccl:exit-key-handler (ccl:current-key-handler view) nil)
  (flet ((insured-time (buf default)
           (multiple-value-bind (val flag) (textbuf-form buf default)
             (if (and flag (numberp val) (>= val 0))
               val
               (progn (warn-user "~A is not a start time."
                                 (textbuf-text buf))
                      (return-from worksheet-action nil))))))
    (let* ((seq-pane (ccl:view-named 'sequence-pane view))
           (start (insured-time (ccl:view-named 'start-typein seq-pane) 0))
           (repeat (insured-time (ccl:view-named 'repeat-typein seq-pane) 1))
           (pause (insured-time (ccl:view-named 'pause-typein seq-pane) 0))
           (objects '()))
      (catch :abort
        (multiple-value-bind (stream syntax type text)
                             (output-spec-to-stream-object
                              (ccl:view-named 'output-pane view))
          (declare (ccl:ignore-if-unused stream))
          (setf objects (get-current-reference 
                         (ccl:view-named 'sequence-pane view)))
          (let ((*syntax* syntax)
                (*command-prompting* nil))
            (if (eq type :file)
              (ccl:with-cursor ccl:*watch-cursor*
                (swrite-object objects text (list start repeat pause)))
              (slisten-object objects (list start repeat pause)))))))))
                   

;;;
;;; -*- EOF -*-