;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :cm)

(defsyntax common-lisp-music
   :pathname (pathname "test.clm")
   :stream-types '((clm-sound-file "snd")
                   (clm-sound-file "aiff")
                   (clm-score-file "clm"))
   :nicknames '(clm))

(defparameter *clm-trace* t)

(defclass snd-mixin ()
  ((channels :initarg channels :initarg :channels :initform 1)
   (srate :initarg srate :initarg :srate :initarg sampling-rate
          :initarg :sampling-rate :initform clm:sampling-rate)
   (continue-old-file :initarg continue-old-file
                      :initarg :continue-old-file)
   (reverb :initarg reverb :initarg :reverb)
   (sndfile :initarg sndfile :initarg :sndfile)
   (data-format :initarg data-format :initarg :data-format)
   (reverb-data :initarg reverb-data :initarg :reverb-data)
   (reverb-channels :initarg reverb-channels :initarg :reverb-channels
                    :initform 1)
   (decay-time :initarg decay-time :initarg :decay-time :initform 1.0)
   (revfile :initarg revfile :initarg :revfile)
   (play-options :initarg play-options :initarg :play-options)
   (wait :initarg wait :initarg :wait)
   (notehook :initarg notehook :initarg :notehook)
   (statistics :initarg statistics :initarg :statistics)
   (info :initarg info :initarg :info)
   (type :initarg type :initarg :type)
   (force-recomputation :initarg force-recomputation
                        :initarg :force-recomputation)
   (verbose :initarg verbose :initarg :verbose
            :initform clm:*clm-verbose*)
   (comment :initarg comment :initarg :comment
            :initarg commentary :initarg commentary)
   (scaled-to :initarg scaled-to :initarg :scaled-to))
  )

(defobject clm-score-file (event-file header-mixin snd-mixin)
  ((syntax :initform (find-syntax ':clm)))
  (:parameters start end timescale channels srate statistics
               scaled-to
               reverb reverb-data reverb-channels revfile
               continue-old-file play play-options
               wait notehook))
   
(defobject clm-sound-file (event-file snd-mixin)
  ((syntax :initform (find-syntax ':clm))
   (element-type :initform '(unsigned-byte 32))
   (trace :initform *clm-trace* :initarg trace :initarg :trace))
  (:parameters start end timescale
               srate channels reverb decay-time reverb-data 
               reverb-channels revfile statistics info type sndfile
               data-format force-recomputation verbose comment
               scaled-to))

(defmethod initialize-instance :after ((stream clm-sound-file)
                                       &rest args)
  (declare (ignore args))
  (setf (slot-value stream 'name)
    (if (slot-boundp stream 'name)
        (namestring
         (merge-pathnames (slot-value stream 'name)
                          default-sound-file-name)))))

(defun tell-snd (file channels srate reverb decay-time reverb-data)
  (tell-user "~&File: ~A~%~
              Channels: ~D~%~
              Srate: ~F~%~
              Reverb: ~A~@[, decay time: ~S ~]~@[, reverb data: ~S ~]~%"
              file
              (floor channels)
              srate
              (or reverb "None")
              (and reverb decay-time)
              (and reverb reverb-data)))

;;; override the main open/close methods for sound files because clm
;;; manages this itself via init-with-sound and finish-with-sound.
;;; overriding means we have to check the modes before opening.

(defmethod open-event-stream ((stream clm-sound-file)
                              &optional (mode ':default) &rest args)
  (declare (ignore args)) ; already set by earlier method
  (unless (or (eq mode ':default) (eq mode ':initialize))
    (let ((args (collecting-slots stream
                  &key channels srate continue-old-file
                  scaled-to sndfile
                  (reverb :print-if :value) reverb-data
                  reverb-channels revfile play play-options
                  wait notehook statistics decay-time
                  info type data-format force-recomputation verbose 
                  comment)))
      (setf (slot-value stream 'stream)
            (apply #'clm:init-with-sound :output (object-name stream)
                   args)))
    (let ((reverb (slot-value-or-default stream 'reverb)))
      (tell-snd (object-name stream)
                (floor (slot-value stream 'channels))
                (slot-value stream 'srate)
                reverb
                (slot-value-or-default stream 'decay-time)
                (slot-value-or-default stream 'reverb-data))))
  stream)
  
(defmethod close-event-stream ((stream clm-sound-file) &optional mode)
  ;; don't play if closing with error or force
  (let ((wsd (slot-value stream 'stream)))
    (when mode (setf (clm::wsdat-play wsd) nil))
    (clm:finish-with-sound wsd))
  (slot-makunbound stream 'stream)
  (when (slot-value stream 'trace) (format t "Done!~&"))
  t)
   
(defmethod initialize-stream-for-processing :after ((stream
                                                     clm-score-file))
  (let ((s (slot-value stream 'stream)))
    (format s ";;; Common Music output: ~A~&" (cm::date-string))
    (when (slot-boundp stream 'header)
      (write-line ";;; Header" s)
      (format s "~A" (slot-value stream 'header))
      (format s "~%;;; End of Header~&"))))

(defmethod post-process-stream ((stream clm-sound-file) inits)
  ;; clm:finish-with-sound does the sound file postprocessing for us.
  (declare (ignore inits))
  nil)

;;;
;;;
;;;

(defmethod syntax-default-pathname ((syntax common-lisp-music)
                                    (stream clm-sound-file))
  default-sound-file-name)

(defmethod play-using-syntax ((syntax common-lisp-music)
                              file &rest args)
  (declare (ignore args))
  (let ((type (pathname-type file)))
    (cond ((string-equal type "SND") (dac file))
          ((string-equal type "AIFF") (dac file))
          ((string-equal type "CLM") (load-using-syntax syntax file)))))

(defmethod load-using-syntax ((syntax common-lisp-music) file &rest pairs)
  (when pairs (setf pairs (canonicalize-pairs pairs nil t ':keyword)))
  (tell-snd (namestring (truename file))
            (pair-value :channels pairs 1)
            (pair-value ':srate pairs 22050)
            (pair-value ':reverb pairs)
            (pair-value ':decay-time pairs 1.0)
            (pair-value ':reverb-data pairs))
  (apply #'clm-load file pairs))

;;;
;;; stella:definstument expands into clm:definstument and a defclass
;;; for the  instrument's note object. slots are defined by using
;;; the lambda-list declaration from the instrument definition.
;;; since no output or printing methods are created  compilation and
;;; loading is fast in pcl.  the trade off is that, output methods are
;;;  not specialized and must be computed dynamically as each event
;;; is processed. but given the fact that clm takes time to crank samples
;;; anyway, the extra time for dynamic output formatting is probably
;;; insignificant. and if not, users can still elect to create output
;;; methods for their definstrument notes by hand.
;;;

(defun clm-printer (value &optional (stream *standard-output*))
  (if (consp value)
    (unless (eq (first value) 'quote)
      (write-char #\' stream))
    (if (floatp value)
      (setf value (/ (the fixnum (round value .001))
                     1000.0))))
  (prin1 value stream))

;;; cm:definstument expands into clm:definstrument+defobject.

(defmacro definstrument (name lambda &body body)
  (let* ((options (and (consp name) (cdr name)))
         (object? (getf options ':instrument-only t)))
    (if object?
      `(progn
         (clm:definstrument ,name ,lambda ., body)
         ,(define-clm-class (if options (first name) name)
            lambda options))
      `(clm:definstrument ,name ,lambda ., body))))

;;; define one base class for all clm notes. cm:definstument
;;; handels the details.

(defclass clm-note (note) 
  ((instrument :initarg instrument :initarg :instrument)))

;;;
;;; output methods need to know if an instrument parameter is a keyword
;;; and so on.  The system currently represents a parameters as just a
;;; lists; the clm parameter is (slot printer) where slot is either
;;; the slot name or for keyword parameters a cons: (slot . keyword)
;;; note that in the case of clm notes, object-parameters returns
;;; a sorted list of lists: (Requireds Optionals Keywords Rest)
;;; a better method would be to return a single list of parameters in
;;; which each parameter has a Type attribute.  Maybe ill fix this
;;; one day.
;;;

;;; paramter-named already defined in kernel/utils2.lisp

(defun parameter-printer (list) (second list))

;;; one print-object handles all clm notes. prints the rhythm
;;; and the required instrument parameters.

(defmethod print-object ((object clm-note) stream)
  (format stream "#<")
  (print-clm-object object stream t)
  (format stream ">"))

(defun print-clm-object (object stream &optional listing)
  (macrolet ((parameter-printer (list) `(second ,list)))
    (let ((pars (object-parameters object)) )
      ;; write left paren, name 
      (write-char (code-char 40) stream)
      (prin1 (slot-value-or-default object 'instrument
                (class-name (class-of object))) stream)
      ;; write required parameters
      (if listing
          (loop for p in (first pars)
                for s = (parameter-name p)
                do
                (when (eq s 'time) (setf s 'rhythm))
                (write-char #\space stream)
                (if (slot-boundp object s)
                    (funcall (parameter-printer p) 
                             (slot-value object s) 
                             stream)
                    (princ +slot-unset+ stream)))
         (dolist (p (first pars))
           (write-char #\space stream)
           (funcall (parameter-printer p)
                    (slot-value object (parameter-name p)) stream)))
      ;; write non null optionals
      (dolist (p (second pars))
        (let ((v (slot-value object (parameter-name p))))
          (when v
            (write-char #\space stream)
            (funcall (parameter-printer p) v stream))))
      ;; write non null &keywords pars preceded by key
      (dolist (p (third pars))
        (let* ((s (parameter-name p))
               (v (slot-value object (car s))))
          (when v
            (write-char #\space stream)
            (prin1 (cdr s) stream)
            (write-char #\space stream)
            (funcall (parameter-printer p) v stream))))
      ;; write non null &rest arg if no &keys
      (let ((p (fourth pars)))
        (when p
          (unless (third pars)
            (dolist (v (slot-value object (parameter-name p)))
              (write-char #\space stream)
              (funcall (parameter-printer p) v stream)))))
      (write-char (code-char 41) stream)
      object)))

(defun define-clm-class (name lambda-list options)
  (flet (
         ;(makepar (slot printer) `(list ',slot (function ,printer)))
         (makepar (slot printer) (list slot (coerce printer 'function)))
         )
    (let ((mode '&required)
          (time 0)
          (printer 'clm-printer)
          requireds slots optionals keys rest)
      ;; process options
      (loop for (p v) on options by #'cddr
            do 
            (case p
              (:printer 
               (setf printer
                     (if (and (consp v) (eq (car v) 'quote))
                       (second v) v)))
              (:time-parameter 
               (setf time 
                     (if (consp v)
                       (if (eq (car v) 'quote)
                         (second v) 
                         (error "Bogus TIME-PARAMETER: ~S" v))
                       v)))))
      (loop with s for p in lambda-list
            do
            (when (consp p) (setf p (car p)))
            (cond ((eq p '&optional) (setf mode p))
                  ((eq p '&rest) (setf mode p))
                  ((eq p '&key)  (setf mode p))
                  ((eq p '&aux) (return))
                  ((eq p '&allow-other-keys) (return))
                  (t
                   (setf s `(:initarg ,p :initarg
                                      ,(find-symbol (symbol-name p)
                                                    :keyword)))
                   (ecase mode
                     (&required
                      (push (list* p s) slots)
                      (push (makepar p printer) requireds))
                     (&optional
                      (push (list* p ':initform nil s) slots)
                      (push (makepar p printer) optionals))
                     (&key
                      (push (list* p ':initform nil s) slots)
                      (push (makepar (cons p (intern (symbol-name p)
                                                     :keyword))
                                     printer)
                            keys))
                     (&rest
                      (push (list* p :initform nil s) slots)
                      (push (makepar p printer) rest))))))
      (setf slots (nreverse slots))
      (setf requireds (nreverse requireds))
      (setf optionals (nreverse optionals))
      (if (cdr rest)
        (error "Bad lambda list, more than one &rest: ~S" lambda-list)
        (setf rest (first rest)))
      (when keys (setf keys (nreverse keys) rest nil))
      ;; now determine if/how to pass time. remove the instument's
      ;; parameter from slots and update the object's parameter to
      ;; use 'time.
      (unless (not time)
        (let ((pars (append requireds optionals keys rest))
              timeslot timepar)
          (setf timeslot
                (if (numberp time)
                  (nth time slots)
                  (find time slots :key #'first
                        :test #'(lambda (x y)
                                  (string= (symbol-name x)
                                           (symbol-name y))))))
          
          (setf timepar
                (loop for slot = (car timeslot)
                      for par in pars
                      when (eq slot
                               (if (consp (first par)) (caar par)
                                   (first par)))
                      return par))
          (unless (and timeslot timepar)
            (error "Can't resolve :time-parameter ~S." time))
          (if (consp (first timepar))
            (setf (caar timepar) 'time)
            (setf (first timepar) 'time))
          (setf slots (delete timeslot slots :test #'eq))))
      `(progn
         (defclass ,name (clm-note) (,@slots) )
         ;; this setf is dangerous - i should fix
         ;; define-object-parameters to allow preparsed parameters.
         ;; and then use defobject.
         (setf (gethash (find-class ',name) *defobjects*)
               '(,requireds ,optionals ,keys ,rest))))))

;;; the two write-event methods use cached parameter information
;;; to process slots according the different types of lambda pars.

(defmethod write-event ((object clm-note) (stream clm-score-file))
  (let ((file (slot-value stream 'stream)))
    (print-clm-object object file)
    (terpri file) ))

(defmethod write-event ((object clm-note) (file clm-sound-file))
  (let ((pars (object-parameters object))
        (args nil))
    (when (slot-value file 'trace)
      (format t "~6,3F..." (slot-value object 'time)))
    ;; collect required
    (setf args (loop for p in (first pars)
                collect (slot-value object (parameter-name p))))
    ;; append non null optionals
    (when (second pars)
      (nconc args (loop for p in (second pars)
                        for v = (slot-value object (parameter-name p))
                        when v collect v)))

    ;; append non null &keywords with key
    (when (third pars)
      (nconc args (loop for p in (third pars)
                        for s = (parameter-name p)
                        for v = (slot-value object (car s))
                        when v collect (cdr s) and collect v)))
    ;; push non null &rest arg if no &keys
    (when (fourth pars)
      (unless (third pars) (nconc args (list (fourth pars)))))
    ;; call instrument
    (apply (slot-value-or-default object 'instrument
                                  (class-name (class-of object)))
           args)))

